
from sklearn import metrics
from sklearn.model_selection import KFold
from jhsp.Data.Preprocessing.codingdata import CodingData

import os
import sys

# 添加包的顶层目录
top_path = os.path.abspath(__file__)
top_path = top_path.split('jhsp')[0]
top_path = os.path.join(top_path, 'jhsp')
sys.path.append(top_path)






class GetModelScores():
    """

    """

    def __init__(self):
        pass

    def GetScoresCross(self,model,xy_cross_val_list):
        """

        :param model:
        :param x:
        :param y:
        :return:
        """

        score_dict = {}
        acc_score_list = []
        pre_score_list = []
        rec_score_list = []
        f1_score_list = []


        for train_x, train_y, test_x, test_y in xy_cross_val_list:


            train_x = x.iloc[train_index,:]
            train_y = y.iloc[train_index,:]
            test_x = x.iloc[test_index,:]
            test_y = y.iloc[test_index,:]


            model.fit(train_x, train_y)
            pre_y = model.predict(test_x)


            acc_score_list.append(metrics.accuracy_score(test_y,pre_y))

            pre_score_list.append(metrics.precision_score(test_y,pre_y, average='macro'))

            rec_score_list.append(metrics.recall_score(test_y,pre_y, average='macro'))

            f1_score_list.append(metrics.f1_score(test_y,pre_y, average='macro'))




        score_dict['accuracy'] = acc_score_list
        score_dict['precision'] = pre_score_list
        score_dict['recall'] = rec_score_list
        score_dict['f1'] = f1_score_list

        return score_dict

    def GetScores(self,model,train_x,train_y,test_x,test_y):

        model.fit(train_x, train_y)
        pre_y = model.predict(test_x)


        score_dict = {}



        acc_score=(metrics.accuracy_score(test_y,pre_y))

        pre_score=(metrics.precision_score(test_y,pre_y, average='macro'))

        rec_score=(metrics.recall_score(test_y,pre_y, average='macro'))

        f1_score=(metrics.f1_score(test_y,pre_y, average='macro'))




        score_dict['accuracy'] = acc_score
        score_dict['precision'] = pre_score
        score_dict['recall'] = rec_score
        score_dict['f1'] = f1_score

        return score_dict

    def GetScoresNN(self,model,test_x_,test_y_):
        test_x = test_x_.copy()
        test_y = test_y_.copy()


        cd = CodingData()

        pre_y = model.predict(test_x)

        pre_y = cd.onehot2one(pre_y)
        test_y = cd.onehot2one(test_y)


        score_dict = {}



        acc_score=(metrics.accuracy_score(test_y,pre_y))

        pre_score=(metrics.precision_score(test_y,pre_y, average='macro'))

        rec_score=(metrics.recall_score(test_y,pre_y, average='macro'))

        f1_score=(metrics.f1_score(test_y,pre_y, average='macro'))




        score_dict['accuracy'] = acc_score
        score_dict['precision'] = pre_score
        score_dict['recall'] = rec_score
        score_dict['f1'] = f1_score

        return score_dict



from math import pow
from pandas import Series

def my_pow(x,y):
    """
    增加对 series等的支持
    :return:
    """
    try:
        list(x)
    except TypeError:    # 如果是一个数，就会类型错误，如果本身就是series或者list 就不会报错
        return pow(x,y)
    else:
        new_x = x.copy()

        for i in range(len(x)):
            new_x[i] = pow(x[i],y)

        # 返回与X数据类型一致的结果

        if isinstance(x,list):
            return new_x

        elif isinstance(x,Series):
            return Series(new_x)


import copy

def LP(fun):
    """
    to isolate parameters
    """
    def new_fun(*args,**kwargs):
        args_ = copy.deepcopy(args)
        kwargs_ = copy.deepcopy(kwargs)
        fun(*args_,**kwargs_)
    return new_fun


import os
import sys
import numpy as np
import pickle
from sklearn.model_selection import KFold
import pandas as pd

# 添加包的顶层目录
top_path = os.path.abspath(__file__)
top_path = top_path.split('jhsp')[0]
top_path = os.path.join(top_path, 'jhsp')
sys.path.append(top_path)

class GetXY():
    """

    """
    def __init__(self):
        pass

    @staticmethod
    def GetCrossXY(x,y,k,random_state=2021,save_path=False):

        kf = KFold(n_splits=k, shuffle=True, random_state=random_state)

        xy_cross_val_list = []

        for train_index,val_index in kf.split(x):

            # 如果y没有进行onehot的话，就只有一列，是array类型。
            if isinstance(y,pd.DataFrame):
                xy_cross_val_list.append((x.iloc[train_index,:],
                                          y.iloc[train_index, :],
                                          x.iloc[val_index, :],
                                          y.iloc[val_index, :],
                                          ))
            else:
                xy_cross_val_list.append((x.iloc[train_index, :],
                                          y[train_index],
                                          x.iloc[val_index, :],
                                          y[val_index],
                                          ))
        if save_path:
            with open(save_path,'wb') as f:
                pickle.dump(xy_cross_val_list,f)

        return xy_cross_val_list

    @staticmethod
    def LoadCrossXY(path):
        with open(path,'rb') as f:
            xy_cross_val_list = pickle.load(f)
            return xy_cross_val_list













import re

def split(ori_str,seps):

    standard_sep = '!@#$%^&'
    pattern = re.compile(str(seps))
    new_str = re.sub(pattern,standard_sep,ori_str)
    split_str_list = new_str.split(standard_sep)
    
    return split_str_list




import tensorflow as tf
import pandas as pd
import sys
import os

# 添加包的顶层目录
top_path = os.path.abspath(__file__)
top_path = top_path.split('jhsp')[0]
top_path = os.path.join(top_path, 'jhsp')
sys.path.append(top_path)

from Sann.models import sann,sanntuner,anntuner
from Sann import getweights,weights22netweights
import tensorflow_addons as tfa



class GetSannModel():
    """
    返回SANN模型和网络结构一样的ANN模型
    """
    def __init__(self,x,y,how,weights):


        self.weights = weights
        self.out_units = y.shape[1]


        if how == 'mull':
            self.final_activation = 'sigmoid'
            self.loss = 'binary_crossentropy'
            self.metrics = 'accuracy'
        if how == 'mulc':
            self.final_activation = 'softmax'
            self.loss = 'categorical_crossentropy'
            self.metrics = 'accuracy'
        if how =='2c':
            self.final_activation = 'sigmoid'
            self.loss = 'binary_crossentropy'
            self.metrics = 'accuracy'

    # 标签还是分类 ，加一个参数
    def GetSann(self,hp):

        model = sanntuner.Model(self.weights, self.final_activation,hp)
        model.compile(loss=self.loss, optimizer='adam', metrics=[self.metrics])

        return model

    def GetAnn(self,hp):
        """
        获得与SANN网络结构一致的ANN
        :return:
        """

        model = anntuner.Model(self.final_activation,self.out_units,hp)
        model.compile(loss=self.loss, optimizer='adam', metrics=[self.metrics])

        return model

    def GetSannmyself(self,hidden_layers,drop_rate):


        weights = weights22netweights.W22NW().Weights2HW(self.weights,hidden_layers,save=False)[0]




        model = sann.Model(weights, self.final_activation,drop_rate)
        model.compile(loss=self.loss, optimizer=tf.keras.optimizers.Adam(0.3), metrics=['accuracy'])

        return model








import pandas as pd


class StandardizeSym():
    def __init__(self,config_path):
        self.config = {}
        data = pd.read_excel(config_path,header=0)
        for row in range(data.shape[0]):
            if not pd.isna(data.loc[row,'标准化症状']):
                self.config[data.loc[row,'原症状']] = ','.join(data.loc[row,'标准化症状'].split('；'))
            else:
                self.config[data.loc[row, '原症状']] = data.loc[row, '原症状']

    def StandardizeSyms(self,syms_list):
        config = self.config.copy()
        fail_syms = set(syms_list) - set(config.keys())


        if fail_syms:
            for i in fail_syms:
                config[i] = i

        standardize_syms_list = [config[sym] for sym in syms_list]
        return standardize_syms_list,list(fail_syms)



class StandardizeDiags():
    def __init__(self, config_path):

        self.all = []
        data = pd.read_excel(config_path, header=0)
        elements_dict = data.to_dict()
        for i_dict in elements_dict.values():
            self.all.extend(list(i_dict.values()))

        self.k = list(data.loc[:,'证基空间'])
        self.w = list(data.loc[:, '证基物质'])
        self.y = list(data.loc[:, '证基运动'])



    def StandardizeDiag(self, diags,sep='|',mode='all'):
        diag_list = str(diags).split(sep)

        if mode == 'all':
            config = self.all.copy()
        elif mode =='k':
            config = self.k.copy()
        elif mode =='w':
            config = self.w.copy()
        elif mode =='y':
            config = self.y.copy()
        else:
            config = self.all.copy()

        temp_list = []


        for i in range(len(diag_list)):
            for element in config:
                element = str(element)
                if element in diag_list[i]:
                    temp_list.append(element)
                    diag_list[i] = diag_list[i].replace(element,'')


        #返回分割后的三基元素列表；剩余的未能成功标准化分割的列表
        return temp_list,diag_list




import os
import sys
import importlib
import shutil
from tensorflow.keras import callbacks

# 添加包的顶层目录
top_path = os.path.abspath(__file__)
top_path = top_path.split('jhsp')[0]
top_path = os.path.join(top_path, 'jhsp')
sys.path.append(top_path)


from sklearn.model_selection import RandomizedSearchCV
from kerastuner.tuners import RandomSearch

class GetBestModel():

    def __init__(self):


        if os.path.exists('config.py'):
            pass
        else:
            abs_path = os.path.abspath(__file__)
            dir = os.path.dirname(abs_path)
            default_config_path = os.path.join(dir,'default_config.py')
            shutil.copyfile(default_config_path,'config.py')

        import config

        self.config = config


    def GetBestTradition(self,model,parameters,x,y):
        """
        :param model_name:
        :param x:
        :param y:
        :return:
        """

        random_search = RandomizedSearchCV(model,
                                           param_distributions=parameters,
                                           n_iter=self.config.random_search_n_iter,
                                           cv=self.config.cross_val_num,
                                           scoring=self.config.scoring,
                                           random_state=self.config.random_num,
                                           n_jobs= self.config.random_search_n_jobs
                                           )

        random_search.fit(x, y)

        best_params = random_search.cv_results_['params'][random_search.best_index_]
        best_model = random_search.best_estimator_

        return best_model,best_params


    def GetBestNNet(self,model_fun,x_train,y_train,x_val,y_val,path_dir,**kwargs):
        """
        超参数列表已经在构建模型的时候内置到模型里面了，所以不用传入，这人是用keras-tuner调参的一个特性，
        也是与传统调参方法不同的地方。
        """

        tuner = RandomSearch(
            model_fun,
            objective=self.config.scoring,  # 优化目标
            max_trials=self.config.max_trials,
            executions_per_trial=self.config.executions_per_trial,
            directory=path_dir,
            # project_name= model_fun.__name__ ,
        )

        _callbacks = [
            callbacks.EarlyStopping(patience=self.config.patience, min_delta=1e-3,monitor='val_accuracy')
        ]

        tuner.search(x_train,y_train,
                     epochs=self.config.epochs_num,
                     callbacks=_callbacks,
                     # validation_data = (x_val,y_val)
                     )
        
       

        return tuner.get_best_hyperparameters(num_trials=1)[0]
















import getopt
import sys
import os
from time import sleep


def load_config(config_file, args_dict, cmd, load_time=1, inplace=False):
    """
    一个用于从命令行获取参数，更新配置的命令。
    config_file:配置文件路径
    args_dict:需要动态加载的配置字典，如 {'user':'-u','port':'-p'}
    cmd:动态加载配置后执行的命令
    load_time: 配置需要被加载的次数，默认为1. django需要加载两次
    inplace：执行命令后是否保留配置的改变。BOOL类型
    """

    args = ':'.join(list(args_dict.values()))
    args += ':'     # 最后一个参数没有’：‘，所以要加上

    opts = getopt.getopt(sys.argv[1:], args)
    opts = {i[0]: i[1] for i in opts[0]}


    new_config_data = ''
    old_config_data = ''
    with open(config_file, 'r', encoding='utf-8') as f:
        for l in f.readlines():

            old_config_data += l

            for key in args_dict.keys():
                if key in l:
                    _l = l.replace(' ', '')
                    if key+'=' in _l and key+'==' not in _l:   # 排除对参数的判断行,且只匹配赋值行
                        l = "{} = '{}'".format(key, opts[args_dict[key]]) + '\n'
                        break

            new_config_data += l

    # 在new_config_data 末尾增加  在本文件追加一段特殊字符的行，以在文件被加载后确保其被访问（python导入模块，模块的访问时间不改变）
    mark = '##*&@'
    end_line = '''\nwith open('{}', 'a') as f: \n\tf.write("{}")\n'''.format(config_file, mark)
    new_config_data += end_line

    with open(config_file, 'w', encoding='utf-8') as f:
        f.write(new_config_data)


    # 在各平台，解决执行os.system(cmd)中，cmd为django，程序阻塞的问题
    if 'win' in sys.platform:
        cmd = 'start ' + cmd
    elif 'linux' in sys.platform:
        cmd = cmd + ' &'


    os.system(cmd)

    for i in range(3):
        sleep(1)
        with open(config_file, 'r', encoding='utf-8') as f:
            if mark*load_time in f.readlines()[-1]:      # 被加载几次，mark就会被写入几次
                break

    if not inplace:
        with open(config_file, 'w', encoding='utf-8') as f:
            f.write(old_config_data)

    return None





import tensorflow as tf
from jhsp.Sann.weights22netweights import W22NW


class Model(tf.keras.Model):
    def __init__(self,ori_weights,final_activation,hp):
        super().__init__()

        self.layer_list = []

        hidden_layers = []
        w22n = W22NW()


        for i in range(hp.Int('num_layers', 1, 5)):
            hidden_layers.append(hp.Int('units_' + str(i),
                                            min_value=32,
                                            max_value=400,
                                            step=16))

        initial_weigh, error = w22n.Weights2HW(ori_weights, hidden_layers, False)

        for i,weights in enumerate(initial_weigh):

            kernel_initializer = tf.constant_initializer(weights)

            if i != len(initial_weigh) -1:   #不是最后一层的时候，激活函数为rule

                exec("self.layer{} = tf.keras.layers.Dense(units=weights.shape[1],activation='relu' ,kernel_initializer=kernel_initializer)".format(i))

            else:
                exec("self.layer{0} = tf.keras.layers.Dense(units=weights.shape[1],activation='{1}',kernel_initializer=kernel_initializer)".format(i,final_activation))

            exec('self.layer_list.append(self.layer{})'.format(i))
            self.layer_list.append(tf.keras.layers.Dropout(hp.Choice('drop_rate', values=[0.1,0.2,0.3,0.4])))

        self.layer_list = self.layer_list[:-1]


    def build(self, input_shape):
        super(Model, self).build(input_shape)

    def call(self, x):

        for layer in  self.layer_list:
            x = layer(x)

        return x





import tensorflow as tf
import pandas as pd


class Model(tf.keras.Model):
    def __init__(self,initial_weigh,final_activation,drop_rate):
        super().__init__()

        self.layer_list = []

        for i,weights in enumerate(initial_weigh):

            kernel_initializer = tf.constant_initializer(weights)

            if i != len(initial_weigh) -1:   #不是最后一层的时候，激活函数为rule

                exec("self.layer{} = tf.keras.layers.Dense(units=weights.shape[1],activation='relu' ,kernel_initializer=kernel_initializer)".format(i))
            else:
                exec("self.layer{0} = tf.keras.layers.Dense(units=weights.shape[1],activation='{1}',kernel_initializer=kernel_initializer)".format(i,final_activation))

            exec('self.layer_list.append(self.layer{})'.format(i))
            self.layer_list.append(tf.keras.layers.Dropout(drop_rate))

        self.layer_list = self.layer_list[:-1]

    def call(self, x):

        for layer in  self.layer_list:
            x = layer(x)

        return x


























# 通过封装 train evaluate 来进行基于pytorch的模型训练和保存。
# 提供输出保存以及绘图
# 自动在当前目录生成config

from shutil import copyfile
import os
import time
import pandas as pd
import numpy as np
import torch



class TorchPipline:
    def __init__(self):

        self.config = self.load_config()

        self.epoch_data = pd.DataFrame(
            data=np.zeros(shape=(20000, len(self.config.epoce_data_columns))),
            columns= self.config.epoce_data_columns
        )

    def __train__(self, model, optimizer, dataloder):
        """
        实例化后改写
        """
        pass

    def __loss__(self,pres, labels):
        """
        实例化后改写
        """
        pass

    def load_config(self):
        """
        讲config文件从库文件复制到当前目录，再从当前目录载入。
        这样可以方便查看和修改config
        """
        cur_path = os.path.join(
            os.getcwd(),
            'config_torchpipeline.py'
        )

        ori_path = os.path.join(
            os.path.dirname(os.path.abspath('__file__')),
            'config_torchpipeline.py'
        )

        # 如果在当前目录没有存在config文件，则复制
        if not os.path.exists(cur_path):
            copyfile(ori_path, cur_path)

        # load config 文件
        from . import config_torchpipeline

        return config_torchpipeline

    def __update__(self, update_data):
        """
        每一epoch 更新一次
        更新内容包括： 学习率、是否提前停止训练、保存训练结果、保存模型等
        依据epoch_data判断更新lr，例如学习率衰减。
        """

        # for param_group in optimizer.param_groups:
        #     param_group["lr"] = lr

    def load_model(self, e):
        pass

    def save(self, model, epoch_data):
        pass

    def __evaluate__(self,model,dataloder):
        pass

    def __get_dataloder__(self, mode):
        pass

    def train(self, model, optimizer, epoch, dataloder, breakpoint=-1, test=False):
        """
        breakpoint = 3, 则加载3的权值，从4开始训练。
        """

        for e in range(epoch):

            t = time.time()

            # 断点之前不训练
            if e <= breakpoint:
                continue

            # 到断点下一个epoch的的时候，载入断点权值
            elif e == breakpoint + 1:
                # 如果断点是-1，则e=0,是第一个epoch，则不载入权值
                if e != 0:
                    model = self.load_model(breakpoint)


            # 训练
            model = self.__train__(model, optimizer, dataloder)



            # 模型结果评估
            train_loss, train_eva = self.__evaluate__(model, self.__get_dataloder__('train'))
            dev_loss, dev_eva = self.__evaluate__(model, self.__get_dataloder__('dev'))

            if test:
                test_loss, test_eva = self.__evaluate__(model, self.__get_dataloder__('test'))
            else:
                test_loss, test_eva = 'None', 'None'



            # 根据训练结果更新参数
            update_data = {
                'epoch': e,
                'train_loss': train_loss,
                'train_eva': train_eva,
                'dev_loss': dev_loss,
                'dev_eva': dev_eva,
                'test_loss': test_loss,
                'test_eva': test_eva,
                'lr': optimizer.param_groups[0]['lr']
            }

            optimizer = self.__update__(update_data)

            # 保存训练结果和模型

            torch.save(model.state_dict(), self.save.format(e))






import tensorflow as tf


class Model(tf.keras.Model):
    def __init__(self,final_activation,out_units,hp):
        super().__init__()

        self.layer_list = []
        hidden_layers = []

        for i in range(hp.Int('num_layers', 1, 5)):
            hidden_layers.append(hp.Int('units_' + str(i),
                                        min_value=32,
                                        max_value=512,
                                        step=32))

        a = hidden_layers

        for i,units in enumerate(hidden_layers):

            if i != len(hidden_layers) -1:   #不是最后一层的时候，激活函数为rule
                exec("self.layer{} = tf.keras.layers.Dense(units=units,activation='relu' )".format(i))
            else:
                exec("self.layer{} = tf.keras.layers.Dense(units=out_units,activation='{}')".format(i,final_activation))

            exec('self.layer_list.append(self.layer{})'.format(i))
            self.layer_list.append(tf.keras.layers.Dropout(hp.Choice('drop_rate', values=[0.1, 0.2, 0.3, 0.4])))

        self.layer_list = self.layer_list[:-1]

    def call(self, x):

        for layer in  self.layer_list:
            x = layer(x)


        return x






import numpy as np
from sklearn.ensemble import RandomForestClassifier
from sklearn.neighbors import KNeighborsClassifier
from sklearn.svm import SVC


def GetTraMandP(model_name):


    # 模型定义
    RF_parameters = {
        'n_estimators': range(20, 800, 10),
        'criterion': ['entropy', 'gini'],
        'max_depth': range(5, 25, 1),
        'max_features': ['auto', 'sqrt', 'log2'],
        'min_samples_split': range(2, 5, 1),
        'random_state': [2021]
    }

    KNN_parameters = {
        'n_neighbors': range(1, 10, 1),
        'weights': ['uniform', 'distance'],
        'algorithm': ['auto', 'ball_tree', 'kd_tree', 'brute']
    }

    SVC_parameters = {
        'C': np.arange(1, 8, 0.2),
        'kernel': ['linear', 'poly', 'rbf', 'sigmoid'],
        'gamma': ['scale', 'auto'],
        'degree': range(1, 11, 1),
        'random_state': [2021]
    }

    # 模型列表
    models_dict = {'RF': RandomForestClassifier(),
                   'KNN': KNeighborsClassifier(),
                   'SVC': SVC()}

    # 模型参数字典
    param_dict = {'RF': RF_parameters,
                      'KNN': KNN_parameters,
                      'SVC': SVC_parameters}

    return models_dict[model_name],param_dict[model_name]






# -*- coding: utf-8 -*-
from functools import wraps
from collections.abc import Iterable
import sys
import math
import warnings
import tqdm


# test multiprocessing decorator
# 带参数的装饰器的实现方法是再包裹一层函数，来返回装饰器：装饰器生成器（最外层函数）-函数生成器（装饰器）—函数（原函数）

def mp(pool_size=5, batch_size=1, how='process', require_return=False):
    """
    多进程任务，抽象一下，就是一系列参数，得到一系列结果，或者不需要结果。
    原函数的可迭代对象必须是第一个参数
    """

    if how == 'process':
        from multiprocessing import Process, Queue
    elif how == 'thread':
        from multiprocessing.dummy import Process, Queue

    def outer_func(func):

        @wraps(func)
        def decorated_func(*args, **kwargs):

            # 限制参数
            if len(args) != 0:
                raise Exception('期待只接受关键字参数，您可能需要先更改函数参数设计')

            iter_num = 0
            iter_key = list(kwargs.keys())[0]
            for key in kwargs.keys():
                if isinstance(kwargs[key], Iterable):
                    iter_num += 1

            if iter_num < 1:
                raise Exception('未接受到可迭代参数，您可能不需要使用本装饰器')

            elif iter_num > 1:
                warnings.warn('接受到{}个可迭代参数，请确保只通过多进程迭代第一个。'.format(iter_num))

            # 解决报错问题，将原函数注册到__main__中
            func_name = decorated_func.__dict__['__wrapped__']
            sys.modules['__main__'].__dict__[func_name] = decorated_func.__dict__['__wrapped__']

            # 初始化相关数据
            q = Queue(pool_size)
            p_list = []

            if require_return:
                outcome = ['之前生成的' for _ in range(len(kwargs[iter_key]))]
            else:
                outcome = None

            def update_outcome(_start, _stop, _batch_outcome):

                if require_return:
                    for index in range(start, stop):
                        outcome[index] = batch_outcome[index - start]

            # 构造batch_kwarg，并生成batch_p
            batch_num = math.ceil(len(kwargs[iter_key]) / batch_size)

            #
            for batch_index in range(batch_num):

                start = batch_index * batch_size
                stop = (batch_index + 1) * batch_size

                # 如果是最后一个batch或唯一一个batch（batch_size>迭代参数数量时）
                if batch_index == batch_num - 1 or stop > len(kwargs[iter_key]):
                    stop = len(kwargs[iter_key])

                batch_kwarg = kwargs.copy()
                batch_kwarg[iter_key] = kwargs[iter_key][start:stop]

                def batch_func(start_, stop_, batch_kwarg_):
                    batch_outcome_ = func(**batch_kwarg_)
                    q.put((start_, stop_, batch_outcome_))

                p = Process(target=batch_func, args=(start, stop, batch_kwarg))

                p_list.append(p)

            pbar = tqdm.tqdm(
                total=len(kwargs[iter_key]),
                desc='pool_size:{}, batch_size:{} 全局进度'.format(pool_size, batch_size)
            )

            for pi in range(len(p_list)):  # 有多少个子进程，就有多少个返回值

                # 进程池未满的时候，启动子进程。
                if pi < pool_size:
                    p_list[pi].start()

                    # 如果进程池未满就运行完了全部子进程,则每有一个进程运行完毕，就更新一次结果，结束循环。
                    if pi == len(p_list) - 1:
                        for pi_ in range(len(p_list)):
                            p_list[pi_].join()
                            start, stop, batch_outcome = q.get()
                            update_outcome(start, stop, batch_outcome)
                            pbar.update(batch_size)
                        break

                    else:
                        continue

                # 阻塞直到取一个结果，表面一个子进程运行完毕。再启动一个子进程
                start, stop, batch_outcome = q.get()
                update_outcome(start, stop, batch_outcome)
                pbar.update(batch_size)
                p_list[pi].start()

            # 在进程池未满的时候，只启动了进程，没有取值，因此，进程池大小个值没有被取到
            for i in range(pool_size):
                start, stop, batch_outcome = q.get()
                update_outcome(start, stop, batch_outcome)
                pbar.update(batch_size)

            if require_return:
                return outcome

        return decorated_func

    return outer_func


if __name__ == '__main__':

    arg3 = [4, 5, 6, 7, 8]

    @mp(1, 1, require_return=True)
    def demo_func(iter_arg):
        """
        原函数
        """
        out = []
        for i in iter_arg:
            # print(i)
            out.append(i)
        return out

    a = demo_func(iter_arg=arg3)
    print(a)



import pandas as pd
import os
import warnings
import sys
import tqdm

# 添加包的顶层目录
top_path = os.path.abspath(__file__)
top_path = top_path.split('jhsp')[0]
top_path = os.path.join(top_path, 'jhsp')
sys.path.append(top_path)

# 因为在进行编码的时候，会对一些汉字描述的特征进行编码，如：轻度编码为1，重度编码为2。
# 在进行DataFrame的筛选时，就会出现比较 汉字str与编码int的情况，因此会出现警告信息。
# 这个警告信息可以忽略。
warnings.filterwarnings(action='ignore',message='elementwise comparison failed')

class CodingData():
    """
    这是一个编码数据的类，包括将one数据转换为onehot数据，将onehot数据转化为one数据，编码分类信息，编码等级信息等。
    本类中所有函数的参数都以列表形式传入。
    """
    def __init__(self):
        pass

    def onehot2one(self,onehot_data_):

        if not isinstance(onehot_data_,pd.DataFrame):
            onehot_data = pd.DataFrame(onehot_data_)
        else:
            onehot_data = onehot_data_
            onehot_data.columns = [i for i in range(onehot_data.shape[1])]


        onehot_data['one@#$%^'] = 0

        for row in range(onehot_data.shape[0]):

            onehot_data.iloc[row,-1] = int(onehot_data.iloc[row,:].idxmax()) + 1


        return  onehot_data['one@#$%^']

    @staticmethod
    def one2onehot(data,column_name):
        """
        这个函数将一列，转化为onehot编码的多列
        :param one_columns:
        :return:
        """

        data = data.copy()

        # 将这个列的值作为转化后onehot列的列名
        columns = set(data.loc[:,column_name].values.flatten())
        new_columns = []

        for column in columns:   # 遍历每一个将要新建的onehot列名

            new_column = column_name + str(column) # 为了便于区分，onehot列名之前要加上原列的列名
            new_columns.append(new_column)

            # 用定义的列名，新建一个onehot列，其值为0
            data.loc[:,new_column] = pd.Series([0 for i in range(data.shape[0])])

            # 筛选出原列中，值为新建列列名的行，在新建列中将该行赋值为1
            data.loc[data[column_name]==column,new_column] = 1

        data.drop(column_name, axis=1, inplace=True)

        return data,data.loc[:,new_columns]

    def grades22(self,regulations):
        """
        本函数是将分级数据转化为二分类数据，参数是转化规则，本质是一个列表，列表元素是元组如：【（‘轻度’，0），（“重度”，1）】

        """
        # 遍历所有列
        for column in self.data.columns.values:
            for item in regulations:
                # 对每一列，遍历规则，改为2分类数据
                self.data.loc[self.data[column] == item[0],[column]] = item[1]

    def coding_y(self,y_column):
        """
        本函数功能是将y进行编码，并将编码关系储存到数据目录下。参数为一个列表：【‘列名’】
        :param y_column:
        :return:
        """
        y_names = set(self.data.loc[:, y_column].values.flatten())

        coded_relationship = []
        for i,y_name in enumerate(y_names):
            self.data.loc[self.data[y_column[0]]==y_name,[y_column[0]]] = i
            coded_relationship.append((i, y_name))

        save_dir = os.path.join(self.save_dir,'coded_relationship.txt')
        with open(save_dir,'w',encoding='utf-8') as f:
            for item in coded_relationship:
                f.write(str(item[0]) + ': ' + item[1] + '\n')

    @staticmethod
    def content2features(data,content_column_name,sep=','):

        # 获取所有的特征列表
        features_list = []
        for row in range(data.shape[0]):
            features_list.extend(data.loc[row,content_column_name].split(sep=sep))

        features_list = list(set(features_list))

        try:
            features_list.remove('')
        except ValueError:
            pass

        # 生cheng特征列，默认值为0
        for feature in features_list:
            data[str(feature)] = 0


        #
        for row in tqdm.tqdm(range(data.shape[0])):
            row_features_list =  data.loc[row, content_column_name].split(sep=sep)
            for row_feature in row_features_list:
                data.loc[row,row_feature] = 1


        del data[content_column_name]


        return data,data.loc[:,features_list],features_list

    def save(self):
        self.data.to_excel(self.save_path,index=False)


    def ML22classification(self,data,content_column_name,sep=','):
        temp_data = data.copy()

        # 获取所有的特征列表
        features_list = []
        for row in range(temp_data.shape[0]):
            features_list.extend(temp_data.loc[row, content_column_name].split(sep=sep))

        features_list = list(set(features_list))

        twoclassification_dict = {}
        for feature in features_list:
            temp_data[str(feature)] = 0
            for row in range(temp_data.shape[0]):
                if feature in temp_data.loc[row, content_column_name]:
                    temp_data.loc[row, str(feature)] = 1

            twoclassification_dict[str(feature)] = temp_data.loc[:,feature]

        return twoclassification_dict


import pandas as pd
import os
import sys

# 添加包的顶层目录
top_path = os.path.abspath(__file__)
top_path = top_path.split('jhsp')[0]
top_path = os.path.join(top_path, 'jhsp')
sys.path.append(top_path)

class DescribingData:
    def __init__(self, old_dir):
        self.data = self.data = pd.read_excel(old_dir, header=0)  # 第一行默认为表头，数据第一行必须有表头
        self.save_dir = os.path.dirname(old_dir)

    def to_medical_his(self,exclusion=None):
        data = self.data.copy()
        save_path = os.path.join(self.save_dir, 'medical_history.xlsx')

        if exclusion == None:
            columns = list(data.columns.values)

        else:
            columns = set(data.columns.values) - set(exclusion)

        for column in columns:
            if {0,1} == set(data.loc[:,column].tolist()):
                data.loc[data[column]==1,[column]] = column
            else:
                data.loc[:,column].astype('str')
                data.loc[:,column] += column

        data.to_excel(save_path)




import os
import sys


def add_top_path(object_file_path,top_path_name):

    top_path = os.path.abspath(object_file_path)

    top_path = top_path.split(top_path_name)[0]
    top_path = os.path.join(top_path,top_path_name)

    sys.path.append(top_path)

    return top_path







from sklearn.decomposition import NMF
import pandas as pd
import numpy as np
import pickle
from functools import reduce
import os
import sys

# 添加包的顶层目录
top_path = os.path.abspath(__file__)
top_path = top_path.split('jhsp')[0]
top_path = os.path.join(top_path, 'jhsp')
sys.path.append(top_path)


class W22NW():

    def __init__(self):

        self.W = None
        self.DealW = None
        self.DealAgr = None
        self.hidden_layers = None
        self.HiddenWeights = None

    @staticmethod
    def __StandardScaler(W,how='go',arg_dict=None):


        if how == 'go':
            # 全局归一化
            mean = W.mean(axis=0).mean()
            std = W.std(axis=0).std()

            new_W = (W - mean) / std

            return new_W,{'mean':mean,'std':std}
        elif how == 'back':
            new_W = W * arg_dict['std'] + arg_dict['mean']
            return new_W

    @staticmethod
    def __ApplyOffset(W,how='go',arg_dict=None):

        if how == 'go':
            non_negative_offset = abs(W.min(axis=0).min())
            W += non_negative_offset
            return W,{'non_negative_offset':non_negative_offset}
        elif how == 'back':
            W -= arg_dict['non_negative_offset']
            return W

    @staticmethod
    def __MatrixFactorization(ori_W,inter_length):
        model = NMF(n_components=inter_length, init='random', random_state=0,max_iter=10000)
        W = model.fit_transform(ori_W)
        H = model.components_
        return W,H

    @staticmethod
    def GetError(W1,W2):

        error = W1 - W2
        error = abs(error).mean(axis=0).mean()
        return error

    def __PipDealWeights(self,W,how='go',DealAgr=None):

        if how == 'go':
            DealAgr_dict = {}

            new_W,SAgr = self.__StandardScaler(W)
            new_W,AAgr = self.__ApplyOffset(new_W)

            DealAgr_dict.update(SAgr)
            DealAgr_dict.update(AAgr)

            return new_W,DealAgr_dict

        if how == 'back':
            new_W = self.__ApplyOffset(W,how,DealAgr)
            new_W = self.__StandardScaler(new_W,how,DealAgr)
            return new_W

    def __GetOriWeights(self,HW,DealAgr):

        W = reduce(lambda x, y: np.matmul(x, y), HW)
        W = self.__PipDealWeights(W,how='back',DealAgr=DealAgr)
        return W

    def HiddenWeights2W(self,hd_path,save=True):

        with open(hd_path,'rb') as f:
            HD = pickle.load(f)

        HW = HD['HiddenWeights']
        DealAgr = HD['DealAgr']

        W = self.__GetOriWeights(HW,DealAgr)

        if save:
            pd.DataFrame(W,dtype=float).to_excel('.\HiddenWeights2W.xlsx')

        return W

    def Weights2HW(self,ori_W,hidden_layers,save=True):

        self.hidden_layers = hidden_layers
        self.ori_W = ori_W
        self.DealW,self.DealAgr = self.__PipDealWeights(ori_W)

        hidden_weights = []
        ori_w = self.DealW.copy()

        for hidden_nodes in self.hidden_layers:
            W,H = self.__MatrixFactorization(ori_w,hidden_nodes)
            hidden_weights.append(W)
            ori_w = H

        hidden_weights.append(H)

        self.HiddenWeights = hidden_weights

        HiddenWeights_DealAgr = {'HiddenWeights': self.HiddenWeights,
         'DealAgr': self.DealAgr
         }
        error = self.GetError(self.ori_W,self.__GetOriWeights(self.HiddenWeights,self.DealAgr))

        if save:
            with open('.\HiddenWeights_DealAgr.pk', 'wb') as f:
                pickle.dump({'HiddenWeights': self.HiddenWeights,
                             'DealAgr': self.DealAgr
                             }, f)

        return hidden_weights,error










# 公共模型参数配置
scoring = 'accuracy'
cross_val_num = 5
metrics_list = ['accuracy', 'precision', 'recall', 'f1']
random_num = 100



# sklearn-RandomizedSearchCV 超参数调优参数配置
random_search_n_iter = 100
random_search_n_jobs = -1


# keras-tuner 超参数调优参数配置
epochs_num = 1000
max_trials = 50
executions_per_trial = 2
patience = 50


import pandas as pd
import os
import sys
import tqdm

# 添加包的顶层目录
top_path = os.path.abspath(__file__)
top_path = top_path.split('jhsp')[0]
top_path = os.path.join(top_path, 'jhsp')
sys.path.append(top_path)



class GetWeights():
    """
    令i∈X为某一特征，J∈L为某一具体证型，S代表全体样本,具体见SANN论文
    :param x,y 为特征与诊断
    :param columns: 列名
    :param hidden_layer:  隐藏层节点数  如 [100,200,50]
    :return: 权值矩阵
    """

    def __init__(self,x,y,columns=None,index=None):

        self.x = x
        self.y = y

        # 去除一个阳性样本都没有的诊断列以避免在计算途中报错。由于columns中的对应列名没有删除，所以在最后返回的时候赋值为0 返回了
        self.drop_j = []
        for j in range(self.y.shape[1]):
            if self.y.iloc[:,j].sum() == 0:
                self.drop_j.append(j)

        self.y = self.y.drop(self.y.columns[self.drop_j],axis=1)



        self.index = index
        self.columns = columns

        self.Dij = {}
        self.Fi = {}
        self.Cj = {}
        self.R = {}

        self.W = [] # 特征数*标签数的一个矩阵，被self. GetW 赋值，类型是 dataframe
        self.WStandard = None

    def __GetD(self,i,j):
        """
        D值代表某一特征对某一证型的原始支持度，取值范围为[-0.5 ,0.5]。
        D值的绝对值越大，代表特征对证型的支持度越大，诊断意义也就越大。
        若D值的取值为正，则代表特征对证型有正向诊断意义；若取值为负，则代表特征对证型有负向诊断意义；
        若取值为0，则代表在诊断证型的时候无需考虑这一特征。
        :param i:   该特征在X矩阵中的列下标
        :param j:   该标签在Y矩阵中的列下标
        :return:  Dij = (Sij - (Sj/2)) / Sj
        """

        Sj = list(self.y.iloc[:,j] >= 1).count(True)

        Sij = list(self.x[self.y.iloc[:,j] >= 1].iloc[:,i] >= 1).count(True)

        Dij = (Sij - (Sj / 2)) / Sj

        return Dij

    def __GetF(self,i):
        """
        F值代表某一特征对所有证型的原始平均支持度，因此将特征对不同证型的支持度纳入了考虑。
        F值得计算与D值类似，但针对的是不同证型的所有患者。
        :param i:
        :return: Fi = (Si - (S / 2)) / S
        """

        S = self.x.shape[0]

        Si = list(self.x.iloc[:,i] >= 1).count(True)

        Fi = (Si - (S / 2)) / S

        return Fi

    def __GetC(self,j):
        """
        C的计算，依赖之前D的值
        C值反映了所有特征对某一证型的平均贡献度（与F值不同，F值代表某一特征对所有证型的平均贡献度），
        取值范围为[-0.5 ,0.5]。如果C值过高，则在进行证型预测时，很容易将其他证型的患者误判为这一证型；
        如果C值过低，则这一证型的患者不易被诊断。
        :param j:
        :return: Cj
        """
        Cj = 0

        for i in range(self.x.shape[1]):
            Cj += self.Dij[str(i) + str(j)]

        return Cj

    def __GetR(self):
        """
        R值代表的是所有特征对所有证型的平均贡献度，取值范围为[-0.5,0.5]。
        :return:
        """
        R = 0
        for i in range(self.x.shape[1]):
            R += self.Fi[str(i)]

        return R

    def __GetWij(self,i,j):
        """

        Wij = ((Dij - Fi) * (1 + abs(R-0.5)))/ (1 + (Cj - R))
        :param i:
        :param j:
        :return:
        """


        Dij = self.Dij[str(i) + str(j)]
        Fi = self.Fi[str(i)]
        Cj = self.Cj[str(j)]
        R = self.R

        Wij = ((Dij - Fi) * (1 + abs(R-0.5)))/ (1 + max(0,(Cj - R)))

        return Wij

    def __GetWi(self,i):
        """
        获取某一特征对所有标签的权值
        :param i:
        :return: x2Y
        """
        Wi = []
        for j in range(self.y.shape[1]):
            Wi.append(self.__GetWij(i,j))

        return Wi

    def GetW(self):

        # 计算D、F并保存,key是str类型的数字
        for i in tqdm.tqdm(range(self.x.shape[1])):

            self.Fi[str(i)] = self.__GetF(i)

            for j in range(self.y.shape[1]):

                self.Dij[str(i) + str(j)] = self.__GetD(i,j)

        # 计算C/R 之所以分开是因为C、R的计算依赖D、F
        for j in tqdm.tqdm(range(self.y.shape[1])):
            self.Cj[str(j)] = self.__GetC(j)
        self.R = self.__GetR()



        for i in tqdm.tqdm(range(self.x.shape[1])):
            self.W.append(self.__GetWi(i))


        # 将可能存在的，在__init__里被删除的列补上，以确保神经网络能够正确调用
        if self.drop_j:
            for i in range(len(self.W)):
                for j in self.drop_j:
                    self.W[i].insert(j,0)


        self.W = pd.DataFrame(data=self.W,index=self.index,columns=self.columns)

        return self.W

    def save(self,path='.\Weights.xlsx'):
        self.W.to_excel(path)





